#pragma once
#include <stdint.h>

#ifdef __cplusplus
extern "C" {
#endif

#define NEORADIO2PRWRLY_CHANNEL1_ENABLE  0x01
#define NEORADIO2PRWRLY_CHANNEL2_ENABLE  0x02
#define NEORADIO2PRWRLY_CHANNEL3_ENABLE  0x04
#define NEORADIO2PRWRLY_CHANNEL4_ENABLE  0x08
#define NEORADIO2PRWRLY_CHANNEL5_ENABLE  0x10
#define NEORADIO2PRWRLY_CHANNEL6_ENABLE  0x20
#define NEORADIO2PRWRLY_CHANNEL7_ENABLE  0x40
#define NEORADIO2PRWRLY_CHANNEL8_ENABLE  0x80

#ifdef __cplusplus
}
#endif
