/* ----------------------------------------------------------------------------
 * This file was automatically generated by SWIG (http://www.swig.org).
 * Version 3.0.12
 *
 * This file is not intended to be easily readable and contains a number of
 * coding conventions designed to improve portability and efficiency. Do not make
 * changes to this file unless you know what you are doing--modify the SWIG
 * interface file instead.
 * ----------------------------------------------------------------------------- */

#ifndef PHP_NEORADIO2_H
#define PHP_NEORADIO2_H

extern zend_module_entry neoradio2_module_entry;
#define phpext_neoradio2_ptr &neoradio2_module_entry

#ifdef PHP_WIN32
# define PHP_NEORADIO2_API __declspec(dllexport)
#else
# define PHP_NEORADIO2_API
#endif

ZEND_NAMED_FUNCTION(_wrap_Neoradio2DeviceInfo_name_set);
ZEND_NAMED_FUNCTION(_wrap_Neoradio2DeviceInfo_name_get);
ZEND_NAMED_FUNCTION(_wrap_Neoradio2DeviceInfo_serial_str_set);
ZEND_NAMED_FUNCTION(_wrap_Neoradio2DeviceInfo_serial_str_get);
ZEND_NAMED_FUNCTION(_wrap_Neoradio2DeviceInfo_vendor_id_set);
ZEND_NAMED_FUNCTION(_wrap_Neoradio2DeviceInfo_vendor_id_get);
ZEND_NAMED_FUNCTION(_wrap_Neoradio2DeviceInfo_product_id_set);
ZEND_NAMED_FUNCTION(_wrap_Neoradio2DeviceInfo_product_id_get);
ZEND_NAMED_FUNCTION(_wrap_Neoradio2DeviceInfo__reserved_set);
ZEND_NAMED_FUNCTION(_wrap_Neoradio2DeviceInfo__reserved_get);
ZEND_NAMED_FUNCTION(_wrap_new_Neoradio2DeviceInfo);
PHP_MINIT_FUNCTION(neoradio2);

#endif /* PHP_NEORADIO2_H */
